@@ -4,7 +4,7 @@ Autoconf-like configuration support.
 """
 
 #
-# Copyright (c) 2001, 2002, 2003, 2004 The SCons Foundation
+# Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007 The SCons Foundation
 #
 # Permission is hereby granted, free of charge, to any person obtaining
 # a copy of this software and associated documentation files (the
@@ -26,7 +26,7 @@ Autoconf-like configuration support.
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 #
 
-__revision__ = "/home/scons/scons/branch.0/branch.96/baseline/src/engine/SCons/SConf.py 0.96.92.D001 2006/04/10 23:13:27 knight"
+__revision__ = "/home/scons/scons/branch.0/baseline/src/engine/SCons/SConf.py 0.97.D001 2007/05/17 11:35:19 knight"
 
 import os
 import re
@@ -39,6 +39,7 @@ import types
 import SCons.Action
 import SCons.Builder
 import SCons.Errors
+import SCons.Job
 import SCons.Node.FS
 import SCons.Taskmaster
 import SCons.Util
@@ -141,7 +142,6 @@ def _createSource( target, source, env ):
     fd.write(source[0].get_contents())
     fd.close()
 def _stringSource( target, source, env ):
-    import string
     return (str(target[0]) + ' <-\n  |' +
             string.replace( source[0].get_contents(),
                             '\n', "\n  |" ) )
@@ -188,6 +188,11 @@ class Streamer:
         Return everything written to orig since the Streamer was created.
         """
         return self.s.getvalue()
+
+    def flush(self):
+        if self.orig:
+            self.orig.flush()
+        self.s.flush()
         
 
 class SConfBuildTask(SCons.Taskmaster.Task):
@@ -229,7 +234,6 @@ class SConfBuildTask(SCons.Taskmaster.Task):
             except AttributeError:
                 # Earlier versions of Python don't have sys.excepthook...
                 def excepthook(type, value, tb):
-                    import traceback
                     traceback.print_tb(tb)
                     print type, value
             apply(excepthook, self.exc_info())
@@ -246,8 +250,7 @@ class SConfBuildTask(SCons.Taskmaster.Task):
         cachable = 1
         for t in self.targets:
             bi = t.get_stored_info()
-            c_bi = isinstance(bi, SConfBuildInfo)
-            if c_bi:
+            if isinstance(bi, SConfBuildInfo):
                 if cache_mode == CACHE:
                     t.set_state(SCons.Node.up_to_date)
                 else:
@@ -480,7 +483,6 @@ class SConf:
             result = self.BuildNodes(nodesToBeBuilt)
 
         finally:
-            # Restor the SPAWN value to the environment.
             self.env['SPAWN'] = save_spawn
 
         _ac_build_counter = _ac_build_counter + 1
@@ -599,7 +601,8 @@ class SConf:
             else:
                 _ac_config_logs[self.logfile] = None
                 log_mode = "w"
-            self.logstream = open(str(self.logfile), log_mode)
+            fp = open(str(self.logfile), log_mode)
+            self.logstream = SCons.Util.Unbuffered(fp)
             # logfile may stay in a build directory, so we tell
             # the build system not to override it with a eventually
             # existing file with the same name in the source directory
@@ -851,11 +854,11 @@ def CheckLib(context, library = None, symbol = "main",
 # Bram: Can only include one header and can't use #ifdef HAVE_HEADER_H.
 
 def CheckLibWithHeader(context, libs, header, language,
-                       call = "main();", autoadd = 1):
+                       call = None, autoadd = 1):
     # ToDo: accept path for library. Support system header files.
     """
     Another (more sophisticated) test for a library.
-    Checks, if library and header is available for language (maybe 'C'
+    Checks, if library and header is available for language (may be 'C'
     or 'CXX'). Call maybe be a valid expression _with_ a trailing ';'.
     As in CheckLib, we support library=None, to test if the call compiles
     without extra link flags.
@@ -868,7 +871,7 @@ def CheckLibWithHeader(context, libs, header, language,
     if not SCons.Util.is_List(libs):
         libs = [libs]
 
-    res = SCons.Conftest.CheckLib(context, libs, "main", prog_prefix,
+    res = SCons.Conftest.CheckLib(context, libs, None, prog_prefix,
             call = call, language = language, autoadd = autoadd)
     context.did_show_result = 1
     return not res
